/* APNG Disassembler 1.4
 *
 * Copyright (c) 2009 Max Stepin
 * maxst at users.sourceforge.net
 *
 * GNU LGPL information
 * --------------------
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <zlib.h>

#if defined(_MSC_VER) && _MSC_VER >= 1300
inline unsigned short swap16(unsigned short data) {return(_byteswap_ushort(data));}
inline unsigned int swap32(unsigned int data) {return(_byteswap_ulong(data));}
#elif __linux__
#include <byteswap.h>
inline unsigned short swap16(unsigned short data) {return(bswap_16(data));}
inline unsigned int swap32(unsigned int data) {return(bswap_32(data));}
#else
inline unsigned short swap16(unsigned short data) {return((data >> 8) | (data << 8));}
inline unsigned int swap32(unsigned int data) {return((swap16(data) << 16) | swap16(data >> 16));}
#endif

#define PNG_ZBUF_SIZE  32768

#define PNG_DISPOSE_OP_NONE        0x00
#define PNG_DISPOSE_OP_BACKGROUND  0x01
#define PNG_DISPOSE_OP_PREVIOUS    0x02

#define PNG_BLEND_OP_SOURCE        0x00
#define PNG_BLEND_OP_OVER          0x01

#define notabc(c) ((c) < 65 || (c) > 122 || ((c) > 90 && (c) < 97))

#define ROWBYTES(pixel_bits, width) \
    ((pixel_bits) >= 8 ? \
    ((width) * (((unsigned int)(pixel_bits)) >> 3)) : \
    (( ((width) * ((unsigned int)(pixel_bits))) + 7) >> 3) )

unsigned char   png_sign[8] = {137, 80, 78, 71, 13, 10, 26, 10};

int mask4[2]={240,15};
int shift4[2]={4,0};

int mask2[4]={192,48,12,3};
int shift2[4]={6,4,2,0};

int mask1[8]={128,64,32,16,8,4,2,1};
int shift1[8]={7,6,5,4,3,2,1,0};

unsigned int    keep_original = 1;
unsigned char   pal[256][3];
unsigned char   trns[256];
unsigned int    palsize, trnssize;
unsigned int    hasTRNS;
unsigned short  trns1, trns2, trns3;
z_stream        zstream;

unsigned int readint(FILE * f1)
{
  unsigned char a, b, c, d;
  fread(&a, 1, 1, f1);
  fread(&b, 1, 1, f1);
  fread(&c, 1, 1, f1);
  fread(&d, 1, 1, f1);
  return ((unsigned int)a<<24)+((unsigned int)b<<16)+((unsigned int)c<<8)+(unsigned int)d;
}

unsigned short readshort(FILE * f1)
{
  unsigned char a, b;
  fread(&a, 1, 1, f1);
  fread(&b, 1, 1, f1);
  return ((unsigned short)a<<8)+(unsigned short)b;
}

unsigned short readshort(unsigned char * p)
{
  return ((unsigned short)(*p)<<8)+(unsigned short)(*(p+1));
}

void read_sub_row(unsigned char * row, unsigned int rowbytes, unsigned int bpp)
{
  unsigned int i;

  for (i=bpp; i<rowbytes; i++)
    row[i] += row[i-bpp];
}

void read_up_row(unsigned char * row, unsigned char * prev_row, unsigned int rowbytes, unsigned int bpp)
{
  unsigned int i;

  if (prev_row)
    for (i=0; i<rowbytes; i++)
      row[i] += prev_row[i];
}

void read_average_row(unsigned char * row, unsigned char * prev_row, unsigned int rowbytes, unsigned int bpp)
{
  unsigned int i;

  if (prev_row)
  {
    for (i=0; i<bpp; i++)
      row[i] += prev_row[i]>>1;
    for (i=bpp; i<rowbytes; i++)
      row[i] += (prev_row[i] + row[i-bpp])>>1;
  } 
  else 
  {
    for (i=bpp; i<rowbytes; i++)
      row[i] += row[i-bpp]>>1;
  }
}

void read_paeth_row(unsigned char * row, unsigned char * prev_row, unsigned int rowbytes, unsigned int bpp)
{
  unsigned int i;
  int a, b, c, pa, pb, pc, p;

  if (prev_row) 
  {
    for (i=0; i<bpp; i++)
      row[i] += prev_row[i];
    for (i=bpp; i<rowbytes; i++)
    {
      a = row[i-bpp];
      b = prev_row[i];
      c = prev_row[i-bpp];
      p = b - c;
      pc = a - c;
      pa = abs(p);
      pb = abs(pc);
      pc = abs(p + pc);
      row[i] += ((pa <= pb && pa <= pc) ? a : (pb <= pc) ? b : c);
    }
  } 
  else 
  {
    for (i=bpp; i<rowbytes; i++)
      row[i] += row[i-bpp];
  }
}

void unpack(unsigned char * dst, unsigned int dst_size, unsigned char * src, unsigned int src_size, unsigned int h, unsigned int rowbytes, unsigned char bpp)
{
  zstream.next_out  = dst;
  zstream.avail_out = dst_size;
  zstream.next_in   = src;
  zstream.avail_in  = src_size;
  inflate(&zstream, Z_FINISH);
  inflateReset(&zstream);

  unsigned char * row = dst;
  unsigned char * prev_row = NULL;

  for (unsigned int j=0; j<h; j++)
  {
    switch (*row++) 
    {
      case 0: break;
      case 1: read_sub_row(row, rowbytes, bpp); break;
      case 2: read_up_row(row, prev_row, rowbytes, bpp); break;
      case 3: read_average_row(row, prev_row, rowbytes, bpp); break;
      case 4: read_paeth_row(row, prev_row, rowbytes, bpp); break;
    }
    prev_row = row;
    row += rowbytes;
  }
}

void compose0(unsigned char * dst1, unsigned int dstbytes1, unsigned char * dst2, unsigned int dstbytes2, unsigned char * src, unsigned int w, unsigned int h, unsigned int bop, unsigned char depth, unsigned int rowbytes)
{
  unsigned int    i, j, g, a;
  unsigned char * sp;
  unsigned char * dp1;
  unsigned int  * dp2;
  unsigned char * row = src;
  unsigned char * out1 = dst1;
  unsigned char * out2 = dst2;

  for (j=0; j<h; j++)
  {
    sp = row+1;
    dp1 = out1;
    dp2 = (unsigned int*)out2;

    if (bop == PNG_BLEND_OP_SOURCE)
    {
      for (i=0; i<w; i++)
      {
        switch (depth)
        {
          case 16: for (i=0; i<w; i++) { a = 0xFF; if (hasTRNS && readshort(sp)==trns1) a = 0; *dp1++ = *sp; *dp2++ = (a << 24) + (*sp << 16) + (*sp << 8) + *sp; sp+=2; }  break;
          case 8:  for (i=0; i<w; i++) { a = 0xFF; if (hasTRNS && *sp==trns1)           a = 0; *dp1++ = *sp; *dp2++ = (a << 24) + (*sp << 16) + (*sp << 8) + *sp; sp++;  }  break;
          case 4:  for (i=0; i<w; i++) { g = (sp[i>>1] & mask4[i&1]) >> shift4[i&1]; a = 0xFF; if (hasTRNS && g==trns1) a = 0; *dp1++ = g*0x11; *dp2++ = (a<<24) + g*0x111111; } break;
          case 2:  for (i=0; i<w; i++) { g = (sp[i>>2] & mask2[i&3]) >> shift2[i&3]; a = 0xFF; if (hasTRNS && g==trns1) a = 0; *dp1++ = g*0x55; *dp2++ = (a<<24) + g*0x555555; } break;
          case 1:  for (i=0; i<w; i++) { g = (sp[i>>3] & mask1[i&7]) >> shift1[i&7]; a = 0xFF; if (hasTRNS && g==trns1) a = 0; *dp1++ = g*0xFF; *dp2++ = (a<<24) + g*0xFFFFFF; } break;
        }
      }
    }
    else /* PNG_BLEND_OP_OVER */
    {
      for (i=0; i<w; i++)
      {
        switch (depth)
        {
          case 16: for (i=0; i<w; i++, dp1++, dp2++) { if (readshort(sp) != trns1) { *dp1 = *sp; *dp2 = 0xFF000000 + (*sp << 16) + (*sp << 8) + *sp; } sp+=2; } break;
          case 8:  for (i=0; i<w; i++, dp1++, dp2++) { if (*sp != trns1)           { *dp1 = *sp; *dp2 = 0xFF000000 + (*sp << 16) + (*sp << 8) + *sp; } sp++;  } break;
          case 4:  for (i=0; i<w; i++, dp1++, dp2++) { g = (sp[i>>1] & mask4[i&1]) >> shift4[i&1]; if (g != trns1) { *dp1 = g*0x11; *dp2 = 0xFF000000+g*0x111111; } } break;
          case 2:  for (i=0; i<w; i++, dp1++, dp2++) { g = (sp[i>>2] & mask2[i&3]) >> shift2[i&3]; if (g != trns1) { *dp1 = g*0x55; *dp2 = 0xFF000000+g*0x555555; } } break;
          case 1:  for (i=0; i<w; i++, dp1++, dp2++) { g = (sp[i>>3] & mask1[i&7]) >> shift1[i&7]; if (g != trns1) { *dp1 = g*0xFF; *dp2 = 0xFF000000+g*0xFFFFFF; } } break;
        }
      }
    }

    row += (rowbytes+1);
    out1 += dstbytes1;
    out2 += dstbytes2;
  }
}

void compose2(unsigned char * dst1, unsigned int dstbytes1, unsigned char * dst2, unsigned int dstbytes2, unsigned char * src, unsigned int w, unsigned int h, unsigned int bop, unsigned char depth, unsigned int rowbytes)
{
  unsigned int    i, j;
  unsigned int    r, g, b, a;
  unsigned char * sp;
  unsigned char * dp1;
  unsigned int  * dp2;
  unsigned char * row = src;
  unsigned char * out1 = dst1;
  unsigned char * out2 = dst2;

  for (j=0; j<h; j++)
  {
    sp = row+1;
    dp1 = out1;
    dp2 = (unsigned int*)out2;

    if (bop == PNG_BLEND_OP_SOURCE)
    {
      if (depth == 8)
      {
        for (i=0; i<w; i++)
        {
          b = *sp++;
          g = *sp++;
          r = *sp++;
          a = 0xFF;
          if (hasTRNS && b==trns1 && g==trns2 && r==trns3) 
            a = 0;
          *dp1++ = b; *dp1++ = g; *dp1++ = r;
          *dp2++ = (a << 24) + (r << 16) + (g << 8) + b;
        }
      }
      else
      {
        for (i=0; i<w; i++, sp+=6)
        {
          b = *sp;
          g = *(sp+2);
          r = *(sp+4);
          a = 0xFF;
          if (hasTRNS && readshort(sp)==trns1 && readshort(sp+2)==trns2 && readshort(sp+4)==trns3)
            a = 0;
          *dp1++ = b; *dp1++ = g; *dp1++ = r;
          *dp2++ = (a << 24) + (r << 16) + (g << 8) + b;
        }
      }
    }
    else /* PNG_BLEND_OP_OVER */
    {
      if (depth == 8)
      {
        for (i=0; i<w; i++, sp+=3, dp1+=3, dp2++)
        if ((*sp != trns1) || (*(sp+1) != trns2) || (*(sp+2) != trns3))
        {
          *dp1 = *sp; *(dp1+1) = *(sp+1); *(dp1+2) = *(sp+2);
          *dp2 = 0xFF000000 + (*(sp+2) << 16) + (*(sp+1) << 8) + *sp;
        }
      }
      else
      {
        for (i=0; i<w; i++, sp+=6, dp1+=3, dp2++)
        if ((readshort(sp) != trns1) || (readshort(sp+2) != trns2) || (readshort(sp+4) != trns3))
        {
          *dp1 = *sp; *(dp1+1) = *(sp+2); *(dp1+2) = *(sp+4);
          *dp2 = 0xFF000000 + (*(sp+4) << 16) + (*(sp+2) << 8) + *sp;
        }
      }
    }
    row += (rowbytes+1);
    out1 += dstbytes1;
    out2 += dstbytes2;
  }
}

void compose3(unsigned char * dst1, unsigned int dstbytes1, unsigned char * dst2, unsigned int dstbytes2, unsigned char * src, unsigned int w, unsigned int h, unsigned int bop, unsigned char depth, unsigned int rowbytes)
{
  unsigned int    i, j;
  unsigned int    r, g, b, a;
  unsigned int    r2, g2, b2, a2;
  int             u, v, al;
  unsigned char   col;
  unsigned char * sp;
  unsigned char * dp1;
  unsigned int  * dp2;
  unsigned char * row = src;
  unsigned char * out1 = dst1;
  unsigned char * out2 = dst2;

  for (j=0; j<h; j++)
  {
    sp = row+1;
    dp1 = out1;
    dp2 = (unsigned int*)out2;

    for (i=0; i<w; i++)
    {
      switch (depth)
      {
        case 8: col = sp[i]; break;
        case 4: col = (sp[i>>1] & mask4[i&1]) >> shift4[i&1]; break;
        case 2: col = (sp[i>>2] & mask2[i&3]) >> shift2[i&3]; break;
        case 1: col = (sp[i>>3] & mask1[i&7]) >> shift1[i&7]; break;
      }

      b = pal[col][0];
      g = pal[col][1];
      r = pal[col][2];
      a = trns[col];

      if (bop == PNG_BLEND_OP_SOURCE)
      {
        *dp1++ = col;
        *dp2++ = (a << 24) + (r << 16) + (g << 8) + b;
      }
      else /* PNG_BLEND_OP_OVER */
      {
        if (a == 255)
        {
          *dp1++ = col;
          *dp2++ = (a << 24) + (r << 16) + (g << 8) + b;
        }
        else
        if (a != 0)
        {
          if (a2 = (*dp2)>>24)
          {
            keep_original = 0;
            u = a*255;
            v = (255-a)*a2;
            al = 255*255-(255-a)*(255-a2);
            b2 = ((*dp2)&255);
            g2 = (((*dp2)>>8)&255);
            r2 = (((*dp2)>>16)&255);
            b = (b*u + b2*v)/al;
            g = (g*u + g2*v)/al;
            r = (r*u + r2*v)/al;
            a = al/255;
          }
          *dp1++ = col;
          *dp2++ = (a << 24) + (r << 16) + (g << 8) + b;
        }
        else
        {
          dp1++;
          dp2++;
        }
      }
    }
    row += (rowbytes+1);
    out1 += dstbytes1;
    out2 += dstbytes2;
  }
}

void compose4(unsigned char * dst, unsigned int dstbytes, unsigned char * src, unsigned int w, unsigned int h, unsigned int bop, unsigned char depth, unsigned int rowbytes)
{
  unsigned int    i, j, step;
  unsigned int    g, a, g2, a2;
  int             u, v, al;
  unsigned char * sp;
  unsigned char * dp;
  unsigned char * row = src;
  unsigned char * out = dst;

  step = (depth+7)/8;

  for (j=0; j<h; j++)
  {
    sp = row+1;
    dp = out;

    if (bop == PNG_BLEND_OP_SOURCE)
    {
      for (i=0; i<w; i++)
      {
        g = *sp; sp += step;
        a = *sp; sp += step;
        *dp++ = g;
        *dp++ = a;
      }
    }
    else /* PNG_BLEND_OP_OVER */
    {
      for (i=0; i<w; i++)
      {
        g = *sp; sp += step;
        a = *sp; sp += step;
        if (a == 255)
        {
          *dp++ = g;
          *dp++ = a;
        }
        else
        if (a != 0)
        {
          if (a2 = *(dp+1))
          {
            u = a*255;
            v = (255-a)*a2;
            al = 255*255-(255-a)*(255-a2);
            g2 = ((*dp)&255);
            g = (g*u + g2*v)/al;
            a = al/255;
          }
          *dp++ = g;
          *dp++ = a;
        }
        else
          dp+=2;
      }
    }
    row += (rowbytes+1);
    out += dstbytes;
  }
}


void compose6(unsigned char * dst, unsigned int dstbytes, unsigned char * src, unsigned int w, unsigned int h, unsigned int bop, unsigned char depth, unsigned int rowbytes)
{
  unsigned int    i, j, step;
  unsigned int    r, g, b, a;
  unsigned int    r2, g2, b2, a2;
  int             u, v, al;
  unsigned char * sp;
  unsigned int  * dp;
  unsigned char * row = src;
  unsigned char * out = dst;

  step = (depth+7)/8;

  for (j=0; j<h; j++)
  {
    sp = row+1;
    dp = (unsigned int*)out;

    if (bop == PNG_BLEND_OP_SOURCE)
    {
      for (i=0; i<w; i++)
      {
        b = *sp; sp += step;
        g = *sp; sp += step;
        r = *sp; sp += step;
        a = *sp; sp += step;
        *dp++ = (a << 24) + (r << 16) + (g << 8) + b;
      }
    }
    else /* PNG_BLEND_OP_OVER */
    {
      for (i=0; i<w; i++)
      {
        b = *sp; sp += step;
        g = *sp; sp += step;
        r = *sp; sp += step;
        a = *sp; sp += step;
        if (a == 255)
          *dp++ = (a << 24) + (r << 16) + (g << 8) + b;
        else
        if (a != 0)
        {
          if (a2 = (*dp)>>24)
          {
            u = a*255;
            v = (255-a)*a2;
            al = 255*255-(255-a)*(255-a2);
            b2 = ((*dp)&255);
            g2 = (((*dp)>>8)&255);
            r2 = (((*dp)>>16)&255);
            b = (b*u + b2*v)/al;
            g = (g*u + g2*v)/al;
            r = (r*u + r2*v)/al;
            a = al/255;
          }
          *dp++ = (a << 24) + (r << 16) + (g << 8) + b;
        }
        else
          dp++;
      }
    }
    row += (rowbytes+1);
    out += dstbytes;
  }
}

void write_chunk(FILE * f, const char * name, unsigned char * data, unsigned int length)
{
  unsigned int crc = crc32(0, Z_NULL, 0);
  unsigned int len = swap32(length);

  fwrite(&len, 1, 4, f);
  fwrite(name, 1, 4, f);
  crc = crc32(crc, (const Bytef *)name, 4);

  if (data != NULL && length > 0)
  {
    fwrite(data, 1, length, f);
    crc = crc32(crc, data, length);
  }

  unsigned int crc2 = swap32(crc);
  fwrite(&crc2, 1, 4, f);
}

void write_IDATs(FILE * f, unsigned char * data, unsigned int length, unsigned int idat_size)
{
  unsigned int z_cmf = data[0];
  if ((z_cmf & 0x0f) == 8 && (z_cmf & 0xf0) <= 0x70)
  {
    if (length >= 2)
    {
      unsigned int z_cinfo = z_cmf >> 4;
      unsigned int half_z_window_size = 1 << (z_cinfo + 7);
      while (idat_size <= half_z_window_size && half_z_window_size >= 256)
      {
        z_cinfo--;
        half_z_window_size >>= 1;
      }
      z_cmf = (z_cmf & 0x0f) | (z_cinfo << 4);
      if (data[0] != (unsigned char)z_cmf)
      {
        data[0] = (unsigned char)z_cmf;
        data[1] &= 0xe0;
        data[1] += (unsigned char)(0x1f - ((z_cmf << 8) + data[1]) % 0x1f);
      }
    }
  }

  while (length > 0)
  {
    unsigned int ds = length;
    if (ds > PNG_ZBUF_SIZE)
      ds = PNG_ZBUF_SIZE;

    write_chunk(f, "IDAT", data, ds);

    data += ds;
    length -= ds;
  }
}

void SavePNG(unsigned char * p, unsigned int w, unsigned int h, unsigned int rowbytes, unsigned char coltype)
{
  struct IHDR 
  {
    unsigned int    mWidth;
    unsigned int    mHeight;
    unsigned char   mDepth;
    unsigned char   mColorType;
    unsigned char   mCompression;
    unsigned char   mFilterMethod;
    unsigned char   mInterlaceMethod;
  } ihdr;

  char            szOut[] = "apngdis.png";
  FILE          * f;
  unsigned int    j;

  unsigned int idat_size = (rowbytes + 1) * h;
  unsigned int zbuf_size = idat_size + ((idat_size + 7) >> 3) + ((idat_size + 63) >> 6) + 11;

  unsigned char * row_buf = (unsigned char *)malloc(rowbytes + 1);
  if (row_buf == NULL)
    return;

  unsigned char * zbuf = (unsigned char *)malloc(zbuf_size);
  if (zbuf == NULL)
    return;

  zstream.data_type = Z_BINARY;
  zstream.zalloc    = Z_NULL;
  zstream.zfree     = Z_NULL;
  zstream.opaque    = Z_NULL;
  deflateInit2(&zstream, Z_BEST_COMPRESSION, 8, 15, 8, Z_DEFAULT_STRATEGY);

  if ((f = fopen(szOut, "wb")) != 0)
  {
    fwrite(png_sign, 1, 8, f);

    ihdr.mWidth            = swap32(w);
    ihdr.mHeight           = swap32(h);
    ihdr.mDepth            = 8;
    ihdr.mColorType        = coltype;
    ihdr.mCompression      = 0;
    ihdr.mFilterMethod     = 0;
    ihdr.mInterlaceMethod  = 0;
    write_chunk(f, "IHDR", (unsigned char *)(&ihdr), 13);

    if (coltype == 3)
    {
      if (palsize > 0)
        write_chunk(f, "PLTE", (unsigned char *)(&pal), palsize*3);

      if (trnssize > 0)
        write_chunk(f, "tRNS", trns, trnssize);
    }
    else
    if (coltype == 2)
    {
      if (hasTRNS)
        write_chunk(f, "tRNS", trns, 6);
    }
    else
    if (coltype == 0)
    {
      if (hasTRNS)
        write_chunk(f, "tRNS", trns, 2);
    }

    row_buf[0] = 0;
    zstream.next_out  = zbuf;
    zstream.avail_out = zbuf_size;
    for (j=0; j<h; j++)
    {
      memcpy(row_buf+1, p + j*rowbytes, rowbytes);
      zstream.next_in  = row_buf;
      zstream.avail_in = rowbytes + 1;
      deflate(&zstream, Z_NO_FLUSH);
    }
    deflate(&zstream, Z_FINISH);

    write_IDATs(f, zbuf, zstream.total_out, idat_size);

    write_chunk(f, "IEND", 0, 0);
    fclose(f);
  }
  else
    printf("Error: can't open the file '%s'\n", szOut);

  deflateEnd(&zstream);
  free(zbuf);
  free(row_buf);
}

int main(int argc, char** argv)
{
  char          * szIn;
  int             res;
  unsigned int    i, j;
  unsigned int    rowbytes; 
  int             imagesize, zbuf_size, zsize, trns_idx;
  unsigned int    len, chunk, crc;
  unsigned int    w, h, seq, w0, h0, x0, y0;
  unsigned int    frames, loops, num_fctl, num_idat;
  unsigned int    outrow1, outrow2, outimg1, outimg2;
  unsigned short  d1, d2;
  unsigned char   c, dop, bop;
  unsigned char   channels, depth, pixeldepth, bpp;
  unsigned char   coltype, compr, filter, interl;
  FILE          * f1;

  printf("\nAPNG Disassembler 1.4\n");

  for (i=0; i<256; i++)
  {
    pal[i][0] = i;
    pal[i][1] = i;
    pal[i][2] = i;
    trns[i] = 255;
  }

  zstream.zalloc = Z_NULL;
  zstream.zfree = Z_NULL;
  zstream.opaque = Z_NULL;
  inflateInit(&zstream);

  frames = 1;
  num_fctl = 0;
  num_idat = 0;
  zsize = 0;
  hasTRNS = 0;
  trns_idx = -1;
  x0 = 0;
  y0 = 0;
  bop = PNG_BLEND_OP_SOURCE;

  if (argc <= 1)
  {
    printf("\nUsage: apngdis.exe apng.png\n");
    return 1;
  }

  szIn = argv[1];

  if ((f1 = fopen(szIn, "rb")) != 0)
  {
    unsigned char sig[8];

    if ((res = fread(sig, 1, 8, f1)) == 8)
    {
      if (memcmp(sig, png_sign, 8) == 0)
      {
        len  = readint(f1);
        chunk = readint(f1);

        if ((len == 13) && (chunk == 0x49484452)) /* IHDR */
        {
          w = w0 = readint(f1);
          h = h0 = readint(f1);
          fread(&depth, 1, 1, f1);
          fread(&coltype, 1, 1, f1);
          fread(&compr, 1, 1, f1);
          fread(&filter, 1, 1, f1);
          fread(&interl, 1, 1, f1);
          crc = readint(f1);

          channels = 1;
          if (coltype == 2)
            channels = 3;
          else
          if (coltype == 4)
            channels = 2;
          else
          if (coltype == 6)
            channels = 4;

          pixeldepth = depth*channels;
          bpp = (pixeldepth + 7) >> 3;
          rowbytes = ROWBYTES(pixeldepth, w);

          imagesize = (rowbytes + 1) * h;
          zbuf_size = imagesize + ((imagesize + 7) >> 3) + ((imagesize + 63) >> 6) + 11;

          /*
           * We'll render into 2 output buffers, first in original coltype,
           * second in RGBA.
           *
           * It's better to try to keep the original coltype, but if dispose/blend
           * operations will make it impossible, then we'll save RGBA version instead.
           */

          outrow1 = w*channels; /* output coltype = input coltype */
          outrow2 = w*4;        /* output coltype = RGBA          */
          outimg1 = h*outrow1;
          outimg2 = h*outrow2;

          unsigned char * pOut1=(unsigned char *)malloc(outimg1);
          unsigned char * pOut2=(unsigned char *)malloc(outimg2);
          unsigned char * pTemp=(unsigned char *)malloc(imagesize);
          unsigned char * pData=(unsigned char *)malloc(zbuf_size);
          unsigned char * pDst1;
          unsigned char * pDst2;

          while ( !feof(f1) )
          {
            len  = readint(f1);
            chunk = readint(f1);

            if (chunk == 0x504C5445) /* PLTE */
            {
              unsigned int col;
              for (i=0; i<len; i++)
              {
                fread(&c, 1, 1, f1);
                col = i/3;
                if (col<256)
                {
                  pal[col][i%3] = c;
                  palsize = col+1;
                }
              }
              crc = readint(f1);
            }
            else
            if (chunk == 0x74524E53) /* tRNS */
            {
              hasTRNS = 1;
              for (i=0; i<len; i++)
              {
                fread(&c, 1, 1, f1);
                if (i<256)
                {
                  trns[i] = c;
                  trnssize = i+1;
                  if (c == 0 && coltype == 3 && trns_idx == -1) 
                    trns_idx = i;
                }
              }
              if (coltype == 0)
              {
                trns1 = readshort(&trns[0]);
                if (depth == 16)
                {
                  trns[1] = trns[0]; trns[0] = 0;
                }
              }
              else
              if (coltype == 2)
              {
                trns1 = readshort(&trns[0]);
                trns2 = readshort(&trns[2]);
                trns3 = readshort(&trns[4]);
                if (depth == 16)
                {
                  trns[1] = trns[0]; trns[0] = 0;
                  trns[3] = trns[2]; trns[2] = 0;
                  trns[5] = trns[4]; trns[4] = 0;
                }
              }
              crc = readint(f1);
            }
            else
            if (chunk == 0x6163544C) /* acTL */
            {
              frames = readint(f1);
              loops  = readint(f1);
              crc = readint(f1);
              if (pOut1)
                free(pOut1);
              if (pOut2)
                free(pOut2);
              pOut1=(unsigned char *)malloc(frames*outimg1);
              pOut2=(unsigned char *)malloc(frames*outimg2);
            }
            else
            if (chunk == 0x6663544C) /* fcTL */
            {
              if ((num_fctl == num_idat) && (num_idat > 0))
              {
                if (dop == PNG_DISPOSE_OP_PREVIOUS)
                {
                  if (coltype!=6)
                    memcpy(pOut1 + num_idat*outimg1, pOut1 + (num_idat-1)*outimg1, outimg1);
                  if (coltype!=4)
                    memcpy(pOut2 + num_idat*outimg2, pOut2 + (num_idat-1)*outimg2, outimg2);
                }

                pDst1 = pOut1 + (y0+(num_idat-1)*h)*outrow1 + x0*channels;
                pDst2 = pOut2 + (y0+(num_idat-1)*h)*outrow2 + x0*4;

                unpack(pTemp, imagesize, pData, zsize, h, rowbytes, bpp);
                switch (coltype)
                {
                  case 0: compose0(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
                  case 2: compose2(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
                  case 3: compose3(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
                  case 4: compose4(pDst1, outrow1,                 pTemp, w0, h0, bop, depth, rowbytes); break;
                  case 6: compose6(                pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
                }

                if (dop != PNG_DISPOSE_OP_PREVIOUS)
                {
                  if (coltype!=6)
                    memcpy(pOut1 + num_idat*outimg1, pOut1 + (num_idat-1)*outimg1, outimg1);
                  if (coltype!=4)
                    memcpy(pOut2 + num_idat*outimg2, pOut2 + (num_idat-1)*outimg2, outimg2);

                  if (dop == PNG_DISPOSE_OP_BACKGROUND)
                  {
                    pDst1 += outimg1;
                    pDst2 += outimg2;

                    for (j=0; j<h0; j++)
                    {
                      switch (coltype)
                      {
                        case 0:  memset(pDst2, 0, w0*4); if (hasTRNS) memset(pDst1, trns[1], w0); else keep_original = 0; break;
                        case 2:  memset(pDst2, 0, w0*4); if (hasTRNS) for (i=0; i<w0; i++) { pDst1[i*3] = trns[1]; pDst1[i*3+1] = trns[3]; pDst1[i*3+2] = trns[5]; } else keep_original = 0; break;
                        case 3:  memset(pDst2, 0, w0*4); if (trns_idx >= 0) memset(pDst1, trns_idx, w0); else keep_original = 0; break;
                        case 4:  memset(pDst1, 0, w0*2); break;
                        case 6:  memset(pDst2, 0, w0*4); break;
                      }
                      pDst1 += outrow1;
                      pDst2 += outrow2;
                    }
                  }
                }
              }
              seq = readint(f1);
              w0  = readint(f1);
              h0  = readint(f1);
              x0  = readint(f1);
              y0  = readint(f1);
              d1  = readshort(f1);
              d2  = readshort(f1);
              fread(&dop, 1, 1, f1);
              fread(&bop, 1, 1, f1);
              crc = readint(f1);

              if (num_fctl == 0)
              {
                bop = PNG_BLEND_OP_SOURCE;
                if (dop == PNG_DISPOSE_OP_PREVIOUS)
                  dop = PNG_DISPOSE_OP_BACKGROUND;
              }

              if (!(coltype & 4) && !(hasTRNS))
                bop = PNG_BLEND_OP_SOURCE;

              rowbytes = ROWBYTES(pixeldepth, w0);
              num_fctl++;
            }
            else
            if (chunk == 0x49444154) /* IDAT */
            {
              if (num_fctl > num_idat)
              {
                zsize = 0;
                num_idat++;
              }
              fread(pData + zsize, 1, len, f1);
              zsize += len;
              crc = readint(f1);
            }
            else
            if (chunk == 0x66644154) /* fdAT */
            {
              seq = readint(f1);
              len -= 4;
              if (num_fctl > num_idat)
              {
                zsize = 0;
                num_idat++;
              }
              fread(pData + zsize, 1, len, f1);
              zsize += len;
              crc = readint(f1);
            }
            else
            if (chunk == 0x49454E44) /* IEND */
            {
              break;
            }
            else
            {
              c = (unsigned char)(chunk>>24);
              if (notabc(c)) break;
              c = (unsigned char)((chunk>>16) & 0xFF);
              if (notabc(c)) break;
              c = (unsigned char)((chunk>>8) & 0xFF);
              if (notabc(c)) break;
              c = (unsigned char)(chunk & 0xFF);
              if (notabc(c)) break;

              fseek( f1, len, SEEK_CUR );
              crc = readint(f1);
            }
          }

          if (num_idat == 0)
            num_idat++;

          pDst1 = pOut1 + (y0+(num_idat-1)*h)*outrow1 + x0*channels;
          pDst2 = pOut2 + (y0+(num_idat-1)*h)*outrow2 + x0*4;

          unpack(pTemp, imagesize, pData, zsize, h, rowbytes, bpp);
          switch (coltype)
          {
            case 0: compose0(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
            case 2: compose2(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
            case 3: compose3(pDst1, outrow1, pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
            case 4: compose4(pDst1, outrow1,                 pTemp, w0, h0, bop, depth, rowbytes); break;
            case 6: compose6(                pDst2, outrow2, pTemp, w0, h0, bop, depth, rowbytes); break;
          }

          if (coltype == 0)
          {
            switch (depth)
            {
              case 4: trns[1] *= 0x11; break;
              case 2: trns[1] *= 0x55; break;
              case 1: trns[1] *= 0xFF; break;
            }
          }

          /* Save at PNG strip */
          if (coltype == 6)
            SavePNG(pOut2, w, num_idat*h, outrow2, coltype);
          else
          if (coltype == 4)
            SavePNG(pOut1, w, num_idat*h, outrow1, coltype);
          else
          if (keep_original)
            SavePNG(pOut1, w, num_idat*h, outrow1, coltype);
          else
            SavePNG(pOut2, w, num_idat*h, outrow2, 6);

          if (pData)
            free(pData);
          if (pTemp)
            free(pTemp);
          if (pOut1)
            free(pOut1);
          if (pOut2)
            free(pOut2);
        }
        else
          printf("IHDR missing\n");
      }
      else
        printf("Error: wrong PNG sig\n");
    }
    else
      printf("Error: can't read the sig\n");

    fclose(f1);
  }
  else
    printf("Error: can't open the file\n");

  inflateEnd(&zstream);

  printf("all done\n");
    
  return 0;
}
